# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta import exceptions as exc
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorErrorHandlingTest(test_base.WorkflowConductorTest):
    def test_workflow_input_error(self):
        wf_def = """
        version: 1.0

        input:
          - xyz: <% result().foobar %>

        tasks:
          task1:
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% result().foobar %>'. ExpressionEvaluationException: "
                    "The current task is not set in the context."
                ),
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)

        self.assertRaises(
            exc.InvalidWorkflowStatusTransition, conductor.request_workflow_status, statuses.RUNNING
        )

        self.assertListEqual(conductor.get_next_tasks(), [])
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_errors)

    def test_workflow_input_seq_ref_error(self):
        wf_def = """
        version: 1.0

        input:
          - x
          - y: <% ctx().x %>
          - z: <% ctx().y.value %>

        tasks:
          task1:
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().y.value %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#value"'
                ),
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)

        self.assertRaises(
            exc.InvalidWorkflowStatusTransition, conductor.request_workflow_status, statuses.RUNNING
        )

        self.assert_next_task(conductor, has_next_task=False)
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_errors)

    def test_workflow_vars_error(self):
        wf_def = """
        version: 1.0

        vars:
          - xyz: <% result().foobar %>

        tasks:
          task1:
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% result().foobar %>'. ExpressionEvaluationException: "
                    "The current task is not set in the context."
                ),
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)

        self.assertRaises(
            exc.InvalidWorkflowStatusTransition, conductor.request_workflow_status, statuses.RUNNING
        )

        self.assert_next_task(conductor, has_next_task=False)
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_errors)

    def test_workflow_vars_seq_ref_error(self):
        wf_def = """
        version: 1.0

        vars:
          - x: 123
          - y: <% ctx().x %>
          - z: <% ctx().y.value %>

        tasks:
          task1:
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().y.value %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#value"'
                ),
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)

        self.assertRaises(
            exc.InvalidWorkflowStatusTransition, conductor.request_workflow_status, statuses.RUNNING
        )

        self.assert_next_task(conductor, has_next_task=False)
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_errors)

    def test_task_transition_criteria_error(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% ctx().foobar.fubar %>
                publish:
                  - var1: 'xyz'
                do: task2
              - when: <% succeeded() %>
                do: task3
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - var2: 123
                do: task3
          task3:
            join: all
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task1",
                "task_transition_id": "task2__t0",
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # The workflow should fail on completion of task1 while evaluating task transition.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

        # There are two transitions in task1. The transition to task3 should be processed.
        staged_tasks = list(filter(lambda x: x["id"] == "task3", conductor.workflow_state.staged))
        self.assertGreater(len(staged_tasks), 0)

        # Since the workflow failed, there should be no next tasks returned.
        self.assert_next_task(conductor, has_next_task=False)

    def test_multiple_task_transition_criteria_errors(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar
          - fubar: foobar

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% ctx().foobar.fubar %>
                publish:
                  - var1: 'xyz'
                do: task3
          task2:
            action: core.noop
            next:
              - when: <% ctx().fubar.foobar %>
                publish:
                  - var2: 123
                do: task3
          task3:
            join: all
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task1",
                "task_transition_id": "task3__t0",
            },
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().fubar.foobar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#foobar"'
                ),
                "route": 0,
                "task_id": "task2",
                "task_transition_id": "task3__t0",
            },
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Manually complete task1 and task2. Although the workflow failed when
        # processing task1, task flow can still be updated for task2.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])
        self.forward_task_statuses(conductor, "task2", [statuses.RUNNING, statuses.SUCCEEDED])

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

        # Since both tasks fail evaluating task transition, task3 should not be staged.
        self.assertNotIn("task3", conductor.workflow_state.staged)

        # Since the workflow failed, there should be no next tasks returned.
        self.assert_next_task(conductor, has_next_task=False)

    def test_task_transition_publish_error(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - var1: <% ctx().foobar.fubar %>
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - var2: 123
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task1",
                "task_transition_id": "task2__t0",
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # The workflow should fail on completion of task1 while evaluating task transition.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)
        self.assertNotIn("task2", conductor.workflow_state.staged)

        # Since the workflow failed, there should be no next tasks returned.
        self.assert_next_task(conductor, has_next_task=False)

    def test_task_transition_publish_seq_ref_error(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - x: 123
                  - y: <% ctx().x %>
                  - z: <% ctx().y.value %>
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                publish:
                  - var2: 123
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().y.value %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#value"'
                ),
                "route": 0,
                "task_id": "task1",
                "task_transition_id": "task2__t0",
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # The workflow should fail on completion of task1 while evaluating task transition.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)
        self.assertNotIn("task2", conductor.workflow_state.staged)

        # Since the workflow failed, there should be no next tasks returned.
        self.assert_next_task(conductor, has_next_task=False)

    def test_get_start_tasks_with_task_action_error(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: <% ctx().foobar.fubar %>
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task1",
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # The get_start_tasks method should not return any tasks.
        self.assert_next_task(conductor, has_next_task=False)

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

    def test_get_start_tasks_via_get_next_tasks_with_task_action_error(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: <% ctx().foobar.fubar %>
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task1",
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # The get_next_tasks method should not return any tasks.
        self.assert_next_task(conductor, has_next_task=False)

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

    def test_get_next_tasks_with_task_action_error(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: <% ctx().foobar.fubar %>
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task2",
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Manually complete task1.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # The get_next_tasks method should not return any tasks.
        self.assert_next_task(conductor, has_next_task=False)

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

    def test_get_start_tasks_with_task_input_error(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: core.noop
            input:
              var_x: <% ctx().foobar.fubar %>
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task1",
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # The get_start_tasks method should not return any tasks.
        self.assert_next_task(conductor, has_next_task=False)

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

    def test_get_start_tasks_via_get_next_tasks_with_task_input_error(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: core.noop
            input:
              var_x: <% ctx().foobar.fubar %>
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task1",
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # The get_next_tasks method should not return any tasks.
        self.assert_next_task(conductor, has_next_task=False)

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

    def test_get_next_tasks_with_task_input_error(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
            input:
              var_x: <% ctx().foobar.fubar %>
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task2",
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Manually complete task1.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # The get_next_tasks method should not return any tasks.
        self.assert_next_task(conductor, has_next_task=False)

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

    def test_get_start_tasks_with_multiple_task_action_and_input_errors(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar
          - fubar: foobar

        tasks:
          task1:
            action: <% ctx().foobar.fubar %>
            next:
              - when: <% succeeded() %>
                do: task3
          task2:
            action: core.noop var_x=<% ctx().fubar.foobar %>
            next:
              - when: <% succeeded() %>
                do: task3
          task3:
            join: all
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task1",
            },
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().fubar.foobar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#foobar"'
                ),
                "route": 0,
                "task_id": "task2",
            },
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # The get_start_tasks method should not return any tasks.
        self.assert_next_task(conductor, has_next_task=False)

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

    def test_get_start_tasks_via_get_next_tasks_with_multiple_task_action_and_input_errors(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar
          - fubar: foobar

        tasks:
          task1:
            action: <% ctx().foobar.fubar %>
            next:
              - when: <% succeeded() %>
                do: task3
          task2:
            action: core.noop var_x=<% ctx().fubar.foobar %>
            next:
              - when: <% succeeded() %>
                do: task3
          task3:
            join: all
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task1",
            },
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().fubar.foobar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#foobar"'
                ),
                "route": 0,
                "task_id": "task2",
            },
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # The get_next_tasks method should not return any tasks.
        self.assert_next_task(conductor, has_next_task=False)

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

    def test_get_next_tasks_with_multiple_task_action_and_input_errors(self):
        wf_def = """
        version: 1.0

        description: A basic branching workflow.

        vars:
          - foobar: fubar
          - fubar: foobar

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task2, task3
          task2:
            action: <% ctx().foobar.fubar %>
          task3:
            action: core.noop var_x=<% ctx().fubar.foobar %>
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().foobar.fubar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#fubar"'
                ),
                "route": 0,
                "task_id": "task2",
            },
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().fubar.foobar %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#foobar"'
                ),
                "route": 0,
                "task_id": "task3",
            },
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Manually complete task1.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # The get_next_tasks method should not return any tasks.
        self.assert_next_task(conductor, has_next_task=False)

        # The workflow should fail with the expected errors.
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        actual_errors = sorted(conductor.errors, key=lambda x: x.get("task_id", None))
        self.assertListEqual(actual_errors, expected_errors)

    def test_workflow_output_error(self):
        wf_def = """
        version: 1.0

        output:
          - xyz: <% result().foobar %>

        tasks:
          task1:
            action: core.noop
        """

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% result().foobar %>'. ExpressionEvaluationException: "
                    "The current task is not set in the context."
                ),
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Manually complete task1.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Render workflow output and checkout workflow status and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_errors)
        self.assertIsNone(conductor.get_workflow_output())

    def test_workflow_output_seq_ref_error(self):
        wf_def = """
        version: 1.0

        output:
          - x: 123
          - y: <% ctx().x %>
          - z: <% ctx().y.value %>

        tasks:
          task1:
            action: core.noop
        """

        expected_output = {"x": 123, "y": 123}

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().y.value %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#value"'
                ),
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Manually complete task1.
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING, statuses.SUCCEEDED])

        # Render workflow output and checkout workflow status and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.FAILED)
        self.assertListEqual(conductor.errors, expected_errors)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)



# -*- coding: utf-8 -*-

# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import unittest

from orquesta.utils import strings as str_util


class StringsTest(unittest.TestCase):
    def test_unescape(self):
        self.assertEqual(str_util.unescape("foobar"), "foobar")
        self.assertEqual(str_util.unescape("\\\\foobar"), "\\foobar")

    def test_unicode(self):
        self.assertEqual(str_util.unicode(123), 123)
        self.assertEqual(str_util.unicode("foobar"), "foobar")
        self.assertEqual(str_util.unicode(u"fubar" if six.PY2 else str("fubar")), "fubar")
        self.assertEqual(str_util.unicode("鐵甲奇俠"), "鐵甲奇俠")
        self.assertEqual(str_util.unicode("\xe9\x90\xb5\xe7\x94\xb2"), "\xe9\x90\xb5\xe7\x94\xb2")

    def test_unicode_force(self):
        self.assertEqual(str_util.unicode(123, force=True), "123")
        self.assertEqual(str_util.unicode(123.45, force=True), "123.45")
        self.assertEqual(str_util.unicode(True, force=True), "True")
        self.assertEqual(str_util.unicode(None, force=True), "None")
        self.assertEqual(str_util.unicode([1, 2, 3], force=True), "[1, 2, 3]")
        self.assertEqual(str_util.unicode({"k": "v"}, force=True), "{'k': 'v'}")



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.composition.mistral import base


class BasicWorkflowComposerTest(base.MistralWorkflowComposerTest):
    def test_sequential(self):
        wf_name = "sequential"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    }
                ],
                [
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task2", condition="on-success"),
                    }
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "task1"}, {"id": "task2"}, {"id": "task3"}],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    }
                ],
                [
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task2", condition="on-success"),
                    }
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_parallel(self):
        wf_name = "parallel"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4"},
                {"id": "task5"},
                {"id": "task6"},
            ],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    }
                ],
                [
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task2", condition="on-success"),
                    }
                ],
                [],
                [
                    {
                        "id": "task5",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task4", condition="on-success"),
                    }
                ],
                [
                    {
                        "id": "task6",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task5", condition="on-success"),
                    }
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4"},
                {"id": "task5"},
                {"id": "task6"},
            ],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    }
                ],
                [
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task2", condition="on-success"),
                    }
                ],
                [],
                [
                    {
                        "id": "task5",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task4", condition="on-success"),
                    }
                ],
                [
                    {
                        "id": "task6",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task5", condition="on-success"),
                    }
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_branching(self):
        wf_name = "branching"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4"},
                {"id": "task5"},
            ],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                    {
                        "id": "task4",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                ],
                [
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task2", condition="on-success"),
                    }
                ],
                [],
                [
                    {
                        "id": "task5",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task4", condition="on-success"),
                    }
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [
                {"id": "task1"},
                {"id": "task2"},
                {"id": "task3"},
                {"id": "task4"},
                {"id": "task5"},
            ],
            "adjacency": [
                [
                    {
                        "id": "task2",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                    {
                        "id": "task4",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task1", condition="on-success"),
                    },
                ],
                [
                    {
                        "id": "task3",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task2", condition="on-success"),
                    }
                ],
                [],
                [
                    {
                        "id": "task5",
                        "key": 0,
                        "criteria": self.compose_seq_expr("task4", condition="on-success"),
                    }
                ],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)

    def test_decision_tree(self):
        wf_name = "decision"

        expected_wf_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "t1"}, {"id": "a"}, {"id": "b"}, {"id": "c"}],
            "adjacency": [
                [
                    {
                        "id": "a",
                        "key": 0,
                        "criteria": self.compose_seq_expr(
                            "t1", condition="on-success", expr="<% ctx().which = 'a' %>"
                        ),
                    },
                    {
                        "id": "b",
                        "key": 0,
                        "criteria": self.compose_seq_expr(
                            "t1", condition="on-success", expr="<% ctx().which = 'b' %>"
                        ),
                    },
                    {
                        "id": "c",
                        "key": 0,
                        "criteria": self.compose_seq_expr(
                            "t1", condition="on-success", expr="<% not ctx().which in list(a, b) %>"
                        ),
                    },
                ],
                [],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_graph(wf_name, expected_wf_graph)

        expected_wf_ex_graph = {
            "directed": True,
            "graph": {},
            "nodes": [{"id": "t1"}, {"id": "a"}, {"id": "b"}, {"id": "c"}],
            "adjacency": [
                [
                    {
                        "id": "a",
                        "key": 0,
                        "criteria": self.compose_seq_expr(
                            "t1", condition="on-success", expr="<% ctx().which = 'a' %>"
                        ),
                    },
                    {
                        "id": "b",
                        "key": 0,
                        "criteria": self.compose_seq_expr(
                            "t1", condition="on-success", expr="<% ctx().which = 'b' %>"
                        ),
                    },
                    {
                        "id": "c",
                        "key": 0,
                        "criteria": self.compose_seq_expr(
                            "t1", condition="on-success", expr="<% not ctx().which in list(a, b) %>"
                        ),
                    },
                ],
                [],
                [],
                [],
            ],
            "multigraph": True,
        }

        self.assert_compose_to_wf_ex_graph(wf_name, expected_wf_ex_graph)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta import conducting
from orquesta import events
from orquesta import exceptions as exc
from orquesta import machines
from orquesta.specs import native as native_specs
from orquesta import statuses


class WorkflowStateMachineTest(unittest.TestCase):
    def _prep_conductor(self, status=None):
        wf_def = """
        version: 1.0

        description: A basic sequential workflow.

        tasks:
          task1:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task2
          task2:
            action: core.noop
            next:
              - when: <% succeeded() %>
                do: task3
          task3:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        conductor = conducting.WorkflowConductor(spec)

        if status:
            conductor.request_workflow_status(status)

        return conductor

    def test_bad_event_type(self):
        conductor = self._prep_conductor(statuses.RUNNING)
        tk_ex_event = events.ExecutionEvent("foobar", statuses.RUNNING)
        setattr(tk_ex_event, "task_id", "task1")

        self.assertRaises(
            exc.InvalidEventType,
            machines.WorkflowStateMachine.process_event,
            conductor.workflow_state,
            tk_ex_event,
        )

    def test_bad_event_name(self):
        conductor = self._prep_conductor(statuses.RUNNING)
        tk_ex_event = events.TaskExecutionEvent("task1", 0, statuses.RUNNING)
        setattr(tk_ex_event, "name", "foobar")

        self.assertRaises(
            exc.InvalidEvent,
            machines.WorkflowStateMachine.process_event,
            conductor.workflow_state,
            tk_ex_event,
        )

    def test_bad_event_status(self):
        self.assertRaises(exc.InvalidStatus, events.TaskExecutionEvent, "task1", 0, "foobar")

    def test_bad_current_workflow_status(self):
        conductor = self._prep_conductor()
        conductor.workflow_state.status = statuses.ABANDONED
        tk_ex_event = events.TaskExecutionEvent("task1", 0, statuses.RUNNING)

        self.assertRaises(
            exc.InvalidWorkflowStatusTransition,
            machines.WorkflowStateMachine.process_event,
            conductor.workflow_state,
            tk_ex_event,
        )

    def test_bad_current_workflow_status_to_event_mapping(self):
        conductor = self._prep_conductor(statuses.REQUESTED)
        tk_ex_event = events.TaskExecutionEvent("task1", 0, statuses.RUNNING)

        # If transition is not supported, then workflow status will not change.
        machines.WorkflowStateMachine.process_event(conductor.workflow_state, tk_ex_event)
        self.assertEqual(conductor.get_workflow_status(), statuses.REQUESTED)

    def test_workflow_status_transition(self):
        conductor = self._prep_conductor(statuses.RUNNING)

        tk_ex_event = events.TaskExecutionEvent("task1", 0, statuses.RUNNING)
        machines.WorkflowStateMachine.process_event(conductor.workflow_state, tk_ex_event)
        self.assertEqual(conductor.get_workflow_status(), statuses.RUNNING)

        tk_ex_event = events.TaskExecutionEvent("task1", 0, statuses.PAUSED)
        machines.WorkflowStateMachine.process_event(conductor.workflow_state, tk_ex_event)
        self.assertEqual(conductor.get_workflow_status(), statuses.PAUSED)


class FailedStateTransitionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(FailedStateTransitionTest, cls).setUpClass()
        statuses.ALL_STATUSES.append("mock")

    @classmethod
    def tearDownClass(cls):
        statuses.ALL_STATUSES.remove("mock")
        super(FailedStateTransitionTest, cls).tearDownClass()

    def test_invalid_old_status(self):
        self.assertRaises(
            exc.InvalidStatus,
            machines.WorkflowStateMachine.is_transition_valid,
            "foobar",
            statuses.REQUESTED,
        )

    def test_invalid_new_status(self):
        self.assertRaises(
            exc.InvalidStatus,
            machines.WorkflowStateMachine.is_transition_valid,
            statuses.UNSET,
            "foobar",
        )

    def test_original_status_not_in_transition_map(self):
        self.assertFalse(machines.WorkflowStateMachine.is_transition_valid("mock", None))


class StateTransitionTest(unittest.TestCase):
    def test_null_statuses(self):
        is_transition_valid = machines.WorkflowStateMachine.is_transition_valid
        self.assertTrue(is_transition_valid(None, None))
        self.assertTrue(is_transition_valid(statuses.UNSET, None))
        self.assertTrue(is_transition_valid(None, statuses.UNSET))
        self.assertTrue(is_transition_valid(statuses.UNSET, statuses.UNSET))

    def test_transition(self):
        cases = [
            (x, y)
            for x in machines.WORKFLOW_STATE_MACHINE_DATA.keys()
            for y in machines.WORKFLOW_STATE_MACHINE_DATA[x].values()
        ]

        for x, y in cases:
            expected = x == y or y in machines.WORKFLOW_STATE_MACHINE_DATA[x].values()
            self.assertEqual(machines.WorkflowStateMachine.is_transition_valid(x, y), expected)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorCancelTest(test_base.WorkflowConductorTest):
    def test_workflow_output(self):
        wf_def = """
        version: 1.0

        output:
          - x: 123
          - y: <% ctx().x %>

        tasks:
          task1:
            action: core.noop
        """

        expected_output = {"x": 123, "y": 123}

        expected_errors = []

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        # Run the workflow and keep it running.
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])

        # Cancels the workflow and complete task1.
        conductor.request_workflow_status(statuses.CANCELING)
        self.forward_task_statuses(conductor, "task1", [statuses.SUCCEEDED])

        # Render workflow output and check workflow status and output.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)
        self.assertListEqual(conductor.errors, expected_errors)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_workflow_output_with_error(self):
        wf_def = """
        version: 1.0

        output:
          - x: 123
          - y: <% ctx().x %>
          - z: <% ctx().y.value %>

        tasks:
          task1:
            action: core.noop
        """

        expected_output = {"x": 123, "y": 123}

        expected_errors = [
            {
                "type": "error",
                "message": (
                    "YaqlEvaluationException: Unable to evaluate expression "
                    "'<% ctx().y.value %>'. NoFunctionRegisteredException: "
                    'Unknown function "#property#value"'
                ),
            }
        ]

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        # Run the workflow and keep it running.
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])

        # Cancels the workflow and complete task1.
        conductor.request_workflow_status(statuses.CANCELING)
        self.forward_task_statuses(conductor, "task1", [statuses.SUCCEEDED])

        # Check workflow status is not changed to failed given the output error.
        conductor.render_workflow_output()
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)
        self.assertListEqual(conductor.errors, expected_errors)
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_cancel_workflow_already_canceling(self):
        wf_def = """
        version: 1.0

        tasks:
          task1:
            action: core.noop
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        # Run the workflow and keep it running.
        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)
        self.forward_task_statuses(conductor, "task1", [statuses.RUNNING])

        # Cancels the workflow and complete task1.
        conductor.request_workflow_status(statuses.CANCELING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)

        # Cancels the workflow again.
        conductor.request_workflow_status(statuses.CANCELING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)
        conductor.request_workflow_status(statuses.CANCELED)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)

        # Complete task1 and check workflow status.
        self.forward_task_statuses(conductor, "task1", [statuses.SUCCEEDED])
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELED)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorWithItemsCancelTest(test_base.WorkflowConductorWithItemsTest):
    def assert_cancel_workflow_with_single_item(self, ac_ex_status, tk_ex_status, wf_ex_status):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
        ]

        # Verify the set of action executions.
        expected_task = self.format_task_item(
            task_name,
            task_route,
            task_ctx,
            conductor.spec.tasks.get_task(task_name),
            actions=task_action_specs,
            items_count=len(task_ctx["xs"]),
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        # Set the item to running status.
        self.forward_task_item_statuses(conductor, task_name, 0, [statuses.RUNNING])

        # Assert that the task is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)

        # Cancel the workflow.
        conductor.request_workflow_status(statuses.CANCELING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.CANCELING)

        # Change status for the item.
        result = task_ctx["xs"][0]
        status_changes = [ac_ex_status]
        self.forward_task_item_statuses(conductor, task_name, 0, status_changes, result=result)

        # Assert task and workflow status.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, tk_ex_status)
        self.assertEqual(conductor.get_workflow_status(), wf_ex_status)

    def test_cancel_workflow_with_single_item(self):
        test_cases = [
            (statuses.RUNNING, statuses.CANCELING, statuses.CANCELING),
            (statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELED),
            (statuses.PENDING, statuses.CANCELED, statuses.CANCELED),
            (statuses.PAUSING, statuses.CANCELING, statuses.CANCELING),
            (statuses.RESUMING, statuses.CANCELING, statuses.CANCELING),
            (statuses.PAUSED, statuses.CANCELED, statuses.CANCELED),
            (statuses.CANCELING, statuses.CANCELING, statuses.CANCELING),
            (statuses.CANCELED, statuses.CANCELED, statuses.CANCELED),
            (statuses.FAILED, statuses.CANCELED, statuses.CANCELED),
            (statuses.EXPIRED, statuses.CANCELED, statuses.CANCELED),
            (statuses.ABANDONED, statuses.CANCELED, statuses.CANCELED),
        ]

        for ac_ex_status, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_single_item(ac_ex_status, tk_ex_status, wf_ex_status)

    def assert_cancel_workflow_with_items(self, ac_ex_statuses, tk_ex_status, wf_ex_status):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        # Verify the set of action executions.
        expected_task = self.format_task_item(
            task_name,
            task_route,
            task_ctx,
            conductor.spec.tasks.get_task(task_name),
            actions=task_action_specs,
            items_count=len(task_ctx["xs"]),
        )

        expected_tasks = [expected_task]
        actual_tasks = conductor.get_next_tasks()
        self.assert_task_list(conductor, actual_tasks, expected_tasks)

        # Set the items to running status.
        for i in range(0, len(ac_ex_statuses)):
            self.forward_task_item_statuses(conductor, task_name, i, [statuses.RUNNING])

        # Assert that the task is running.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.RUNNING)

        # Cancel the workflow.
        conductor.request_workflow_status(statuses.CANCELING)
        self.assertEqual(conductor.get_workflow_status(), statuses.CANCELING)
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, statuses.CANCELING)

        # Change status for all but one item.
        for i in range(0, len(ac_ex_statuses) - 1):
            result = task_ctx["xs"][i]
            status_changes = [ac_ex_statuses[i]]
            self.forward_task_item_statuses(conductor, task_name, i, status_changes, result=result)
            actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
            self.assertEqual(actual_task_status, statuses.CANCELING)

        # Change status for the last item.
        i = len(ac_ex_statuses) - 1
        result = task_ctx["xs"][i]
        status_changes = [ac_ex_statuses[i]]
        self.forward_task_item_statuses(conductor, task_name, i, status_changes, result=result)

        # Assert task and workflow status.
        actual_task_status = conductor.workflow_state.get_task(task_name, task_route)["status"]
        self.assertEqual(actual_task_status, tk_ex_status)
        self.assertEqual(conductor.get_workflow_status(), wf_ex_status)

    def test_cancel_workflow_with_items_action_running(self):
        test_cases = [
            # ACTION_RUNNING (rest of the list incomplete)
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RUNNING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RUNNING (not all items processed, rest of the list completed)
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RUNNING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RUNNING (rest of the list completed)
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RUNNING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RUNNING (has item that is paused)
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RUNNING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RUNNING (has item that is canceled)
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RUNNING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RUNNING (has item that failed)
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RUNNING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RUNNING (has item that is expired)
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RUNNING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RUNNING (has item that is abandoned)
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RUNNING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)

    def test_cancel_workflow_with_items_action_pausing(self):
        test_cases = [
            # ACTION_PAUSING (rest of the list incomplete)
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_PAUSING (not all items processed, rest of the list completed)
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_PAUSING (rest of the list completed)
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_PAUSING (has item that is paused)
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_PAUSING (has item that is canceled)
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_PAUSING (has item that failed)
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_PAUSING (has item that is expired)
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_PAUSING (has item that is abandoned)
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)

    def test_cancel_workflow_with_items_action_resuming(self):
        test_cases = [
            # ACTION_RESUMING (rest of the list incomplete)
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RESUMING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RESUMING (not all items processed, rest of the list completed)
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RESUMING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RESUMING (rest of the list completed)
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RESUMING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RESUMING (has item that is paused)
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RESUMING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RESUMING (has item that is canceled)
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RESUMING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RESUMING (has item that failed)
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RESUMING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RESUMING (has item that is expired)
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RESUMING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_RESUMING (has item that is abandoned)
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.RESUMING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)

    def test_cancel_workflow_with_items_action_canceling(self):
        test_cases = [
            # ACTION_CANCELING (rest of the list incomplete)
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_CANCELING (not all items processed, rest of the list completed)
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_CANCELING (rest of the list completed)
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_CANCELING (has item that is paused)
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_CANCELING (has item that is canceled)
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_CANCELING (has item that failed)
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_CANCELING (has item that is expired)
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_CANCELING (has item that is abandoned)
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)

    def test_cancel_workflow_with_items_action_pending(self):
        test_cases = [
            # ACTION_PAUSED_TASK_ACTIVE_ITEMS_INCOMPLETE
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PENDING],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_INCOMPLETE
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PENDING],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_COMPLETED
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PENDING],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_PAUSED
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PENDING],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_CANCELED
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PENDING],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PENDING],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PENDING],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PENDING],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)

    def test_cancel_workflow_with_items_action_paused(self):
        test_cases = [
            # ACTION_PAUSED_TASK_ACTIVE_ITEMS_INCOMPLETE
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSED],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_INCOMPLETE
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_COMPLETED
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_PAUSED
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_CANCELED
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_PAUSED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.PAUSED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)

    def test_cancel_workflow_with_items_action_succeeded(self):
        test_cases = [
            # ACTION_SUCCEEDED_TASK_ACTIVE_ITEMS_INCOMPLETE
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_INCOMPLETE
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_COMPLETED
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED],
                statuses.SUCCEEDED,
                statuses.CANCELED,
            ),
            # ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_PAUSED
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_CANCELED
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_SUCCEEDED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)

    def test_cancel_workflow_with_items_action_failed(self):
        test_cases = [
            # ACTION_FAILED_TASK_ACTIVE_ITEMS_INCOMPLETE
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.FAILED],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_FAILED_TASK_DORMANT_ITEMS_INCOMPLETE
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.FAILED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_FAILED_TASK_DORMANT_ITEMS_COMPLETED
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.FAILED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_FAILED_TASK_DORMANT_ITEMS_PAUSED
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.FAILED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_FAILED_TASK_DORMANT_ITEMS_CANCELED
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.FAILED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_FAILED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.FAILED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_FAILED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.FAILED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_FAILED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.FAILED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)

    def test_cancel_workflow_with_items_action_expired(self):
        test_cases = [
            # ACTION_EXPIRED_TASK_ACTIVE_ITEMS_INCOMPLETE
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.EXPIRED],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_EXPIRED_TASK_DORMANT_ITEMS_INCOMPLETE
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.EXPIRED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_EXPIRED_TASK_DORMANT_ITEMS_COMPLETED
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.EXPIRED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_EXPIRED_TASK_DORMANT_ITEMS_PAUSED
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.EXPIRED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_EXPIRED_TASK_DORMANT_ITEMS_CANCELED
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.EXPIRED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_EXPIRED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.EXPIRED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_EXPIRED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.EXPIRED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_EXPIRED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.EXPIRED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)

    def test_cancel_workflow_with_items_action_abandoned(self):
        test_cases = [
            # ACTION_ABANDONED_TASK_ACTIVE_ITEMS_INCOMPLETE
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.ABANDONED],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_ABANDONED_TASK_DORMANT_ITEMS_INCOMPLETE
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.ABANDONED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_ABANDONED_TASK_DORMANT_ITEMS_COMPLETED
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.ABANDONED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_ABANDONED_TASK_DORMANT_ITEMS_PAUSED
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.ABANDONED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_ABANDONED_TASK_DORMANT_ITEMS_CANCELED
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.ABANDONED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_ABANDONED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.ABANDONED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_ABANDONED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.ABANDONED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_ABANDONED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.ABANDONED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)

    def test_cancel_workflow_with_items_action_canceled(self):
        test_cases = [
            # ACTION_CANCELED_TASK_ACTIVE_ITEMS_INCOMPLETE
            (
                [statuses.RUNNING, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELED],
                statuses.CANCELING,
                statuses.CANCELING,
            ),
            # ACTION_CANCELED_TASK_DORMANT_ITEMS_INCOMPLETE
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_CANCELED_TASK_DORMANT_ITEMS_COMPLETED
            (
                [statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_CANCELED_TASK_DORMANT_ITEMS_PAUSED
            (
                [statuses.PAUSED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_CANCELED_TASK_DORMANT_ITEMS_CANCELED
            (
                [statuses.CANCELED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_CANCELED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.FAILED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_CANCELED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.EXPIRED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
            # ACTION_CANCELED_TASK_DORMANT_ITEMS_FAILED
            (
                [statuses.ABANDONED, statuses.SUCCEEDED, statuses.SUCCEEDED, statuses.CANCELED],
                statuses.CANCELED,
                statuses.CANCELED,
            ),
        ]

        for ac_ex_statuses, tk_ex_status, wf_ex_status in test_cases:
            self.assert_cancel_workflow_with_items(ac_ex_statuses, tk_ex_status, wf_ex_status)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta.tests.unit.conducting.native import base


class BasicWorkflowConductorTest(base.OrchestraWorkflowConductorTest):
    def test_sequential(self):
        wf_name = "sequential"

        expected_task_seq = ["task1", "task2", "task3", "continue"]

        mock_results = [
            "Stanley",
            "All your base are belong to us!",
            "Stanley, All your base are belong to us!",
        ]

        expected_output = {"greeting": mock_results[2]}

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(
            wf_name,
            expected_task_seq,
            inputs={"name": "Stanley"},
            mock_results=mock_results,
            expected_output=expected_output,
        )

    def test_parallel(self):
        wf_name = "parallel"

        expected_task_seq = ["task1", "task4", "task2", "task5", "task3", "task6"]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(wf_name, expected_task_seq)

    def test_branching(self):
        wf_name = "branching"

        expected_task_seq = ["task1", "task2", "task4", "task3", "task5"]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(wf_name, expected_task_seq)

    def test_decision_tree(self):
        wf_name = "decision"

        # Test branch "a"
        expected_task_seq = ["t1", "a"]

        self.assert_conducting_sequences(wf_name, expected_task_seq, inputs={"which": "a"})

        # Test branch "b"
        expected_task_seq = ["t1", "b"]

        self.assert_conducting_sequences(wf_name, expected_task_seq, inputs={"which": "b"})

        # Test branch "c"
        expected_task_seq = ["t1", "c"]

        self.assert_spec_inspection(wf_name)

        self.assert_conducting_sequences(wf_name, expected_task_seq, inputs={"which": "c"})



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta import exceptions as exc
from orquesta.expressions import base as expr_base


class JinjaFacadeEvaluationTest(unittest.TestCase):
    def test_basic_eval(self):
        data = {"foo": "bar"}

        self.assertEqual("bar", expr_base.evaluate("{{ ctx().foo }}", data))
        self.assertEqual("foobar", expr_base.evaluate("foo{{ ctx().foo }}", data))

    def test_basic_eval_undefined(self):
        expr = "{{ ctx().foo }}"

        data = {}

        self.assertRaises(exc.ExpressionEvaluationException, expr_base.evaluate, expr, data)

    def test_dict_eval(self):
        expr = "{{ ctx().nested.foo }}"

        data = {"nested": {"foo": "bar"}}

        self.assertEqual("bar", expr_base.evaluate(expr, data))

    def test_multi_eval(self):
        expr = "{{ ctx().foo }} and {{ ctx().marco }}"

        data = {"foo": "bar", "marco": "polo"}

        self.assertEqual("bar and polo", expr_base.evaluate(expr, data))

    def test_eval_recursive(self):
        expr = "{{ ctx().fee }}"

        data = {
            "fee": "{{ ctx().fi }}",
            "fi": "{{ ctx().fo }}",
            "fo": "{{ ctx().fum }}",
            "fum": "fee-fi-fo-fum",
        }

        self.assertEqual("fee-fi-fo-fum", expr_base.evaluate(expr, data))

    def test_eval_recursive_undefined(self):
        expr = "{{ ctx().fee }}"

        data = {"fee": "{{ ctx().fi }}", "fi": "{{ ctx().fo }}", "fo": "{{ ctx().fum }}"}

        self.assertRaises(exc.ExpressionEvaluationException, expr_base.evaluate, expr, data)

    def test_multi_eval_recursive(self):
        expr = "{{ ctx().fee }} {{ ctx().im }}"

        data = {
            "fee": "{{ ctx().fi }}",
            "fi": "{{ ctx().fo }}",
            "fo": "{{ ctx().fum }}",
            "fum": "fee-fi-fo-fum!",
            "im": "{{ ctx().hungry }}",
            "hungry": "i'm hungry!",
        }

        self.assertEqual("fee-fi-fo-fum! i'm hungry!", expr_base.evaluate(expr, data))

    def test_eval_list(self):
        expr = ["{{ ctx().foo }}", "{{ ctx().marco }}", "foo{{ ctx().foo }}"]

        data = {"foo": "bar", "marco": "polo"}

        self.assertListEqual(["bar", "polo", "foobar"], expr_base.evaluate(expr, data))

    def test_eval_list_of_list(self):
        expr = [["{{ ctx().foo }}", "{{ ctx().marco }}"]]

        data = {"foo": "bar", "marco": "polo"}

        expected = [["bar", "polo"]]

        self.assertListEqual(expected, expr_base.evaluate(expr, data))

    def test_eval_list_of_dict(self):
        expr = [{"foo": "{{ ctx().bar }}", "{{ ctx().marco }}": "polo"}]

        data = {"bar": "bar", "marco": "marco"}

        expected = [{"foo": "bar", "marco": "polo"}]

        self.assertListEqual(expected, expr_base.evaluate(expr, data))

    def test_eval_dict(self):
        expr = {
            "foo": "{{ ctx().bar }}",
            "{{ ctx().marco }}": "polo",
            "foobar": "foo{{ ctx().bar }}",
        }

        data = {"bar": "bar", "marco": "marco"}

        expected = {"foo": "bar", "marco": "polo", "foobar": "foobar"}

        self.assertDictEqual(expected, expr_base.evaluate(expr, data))

    def test_eval_dict_of_dict(self):
        expr = {"nested": {"foo": "{{ ctx().bar }}", "{{ ctx().marco }}": "polo"}}

        data = {"bar": "bar", "marco": "marco"}

        expected = {"nested": {"foo": "bar", "marco": "polo"}}

        self.assertDictEqual(expected, expr_base.evaluate(expr, data))

    def test_eval_dict_of_list(self):
        expr = {"nested": ["{{ ctx().foo }}", "{{ ctx().marco }}"]}

        data = {"foo": "bar", "marco": "polo"}

        expected = {"nested": ["bar", "polo"]}

        self.assertDictEqual(expected, expr_base.evaluate(expr, data))

    def test_type_preservation(self):
        data = {"k1": 101, "k2": 1.999, "k3": True, "k4": [1, 2], "k5": {"k": "v"}, "k6": None}

        self.assertEqual(data["k1"], expr_base.evaluate("{{ ctx().k1 }}", data))

        self.assertEqual(data["k2"], expr_base.evaluate("{{ ctx().k2 }}", data))

        self.assertTrue(expr_base.evaluate("{{ ctx().k3 }}", data))

        self.assertListEqual(data["k4"], expr_base.evaluate("{{ ctx().k4 }}", data))

        self.assertDictEqual(data["k5"], expr_base.evaluate("{{ ctx().k5 }}", data))

        self.assertIsNone(expr_base.evaluate("{{ ctx().k6 }}", data))

    def test_type_string_detection(self):
        expr = "{{ ctx().foo }} -> {{ ctx().bar }}"

        data = {"foo": 101, "bar": 201}

        self.assertEqual("101 -> 201", expr_base.evaluate(expr, data))

    def test_custom_function(self):
        expr = "{{ json('{\"a\": 123}') }}"

        self.assertDictEqual({"a": 123}, expr_base.evaluate(expr))

    def test_custom_function_failure(self):
        expr = "{{ json(int(123)) }}"

        self.assertRaises(exc.ExpressionEvaluationException, expr_base.evaluate, expr)

    def test_block_eval(self):
        expr = "{% for i in ctx().x %}{{ i }}{% endfor %}"

        data = {"x": ["a", "b", "c"]}

        self.assertEqual("abc", expr_base.evaluate(expr, data))

    def test_block_eval_undefined(self):
        expr = "{% for i in ctx().x %}{{ ctx().y }}{% endfor %}"

        data = {"x": ["a", "b", "c"]}

        self.assertRaises(exc.ExpressionEvaluationException, expr_base.evaluate, expr, data)

    def test_block_eval_recursive(self):
        expr = "{% for i in ctx().x %}{{ i }}{% endfor %}"

        data = {
            "x": [
                "{{ ctx().a }}",
                "{{ ctx().b }}",
                "{{ ctx().c }}" "{% for j in ctx().y %}{{ j }}{% endfor %}",
            ],
            "a": "a",
            "b": "b",
            "c": "c",
            "y": ["d", "e", "f"],
        }

        self.assertEqual("abcdef", expr_base.evaluate(expr, data))

    def test_multi_block_eval(self):
        expr = (
            "{% for i in ctx().x %}{{ i }}{% endfor %}" "{% for i in ctx().y %}{{ i }}{% endfor %}"
        )

        data = {"x": ["a", "b", "c"], "y": ["d", "e", "f"]}

        self.assertEqual("abcdef", expr_base.evaluate(expr, data))

    def test_mix_block_and_expr_eval(self):
        expr = "{{ ctx().a }}{% for i in ctx().x %}{{ i }}{% endfor %}{{ ctx().d }}"

        data = {"x": ["b", "c"], "a": "a", "d": "d"}

        self.assertEqual("abcd", expr_base.evaluate(expr, data))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from orquesta.expressions import base as expr_base
from orquesta.expressions.functions import common as core_funcs
from orquesta.expressions import jinja as jinja_expr
from orquesta.expressions import yql as yaql_expr


class ExpressionEvaluatorTest(unittest.TestCase):
    def test_mix_types(self):
        expr = "<% ctx().foo %> and {{ ctx().foo }}"

        expected_errors = [
            {
                "type": None,
                "expression": expr,
                "message": "Expression with multiple types is not supported.",
            }
        ]

        result = expr_base.validate(expr)

        self.assertListEqual(expected_errors, result.get("errors", []))

    def test_inspect_function_has_context_argument(self):
        self.assertTrue(expr_base.func_has_ctx_arg(core_funcs.ctx_))
        self.assertFalse(expr_base.func_has_ctx_arg(core_funcs.json_))

    def test_get_statement_regexes(self):
        expected_data = {
            "jinja": jinja_expr.JinjaEvaluator.get_statement_regex(),
            "yaql": yaql_expr.YAQLEvaluator.get_statement_regex(),
        }

        self.assertDictEqual(expr_base.get_statement_regexes(), expected_data)

    def test_has_expressions(self):
        self.assertTrue(expr_base.has_expressions("<% ctx().foo %> and {{ ctx().foo }}"))
        self.assertTrue(expr_base.has_expressions("foo <% ctx().foo %> bar"))
        self.assertTrue(expr_base.has_expressions("foo {{ ctx().foo }} bar"))
        self.assertFalse(expr_base.has_expressions("foobar"))



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import unittest

from orquesta import conducting
from orquesta import statuses

MOCK_WORKFLOW_STATE = {
    "contexts": [],
    "routes": [],
    "sequence": [],
    "staged": [],
    "status": statuses.UNSET,
    "tasks": {},
}


class WorkflowStateTest(unittest.TestCase):
    def test_get_tasks(self):
        data = copy.deepcopy(MOCK_WORKFLOW_STATE)

        task_sequence = [
            {"id": "task1", "route": 0},
            {"id": "task2", "route": 0},
            {"id": "task3", "route": 0},
            {"id": "task4", "route": 0},
            {"id": "task5", "route": 0},
        ]

        data["sequence"] = copy.deepcopy(task_sequence)
        state = conducting.WorkflowState.deserialize(data)

        actual_task_sequence = state.get_tasks(last_occurrence=False)

        self.assertListEqual(actual_task_sequence, list(enumerate(task_sequence)))

    def test_get_tasks_by_task_id(self):
        data = copy.deepcopy(MOCK_WORKFLOW_STATE)

        task_sequence = [
            {"id": "task1", "route": 0},
            {"id": "task2", "route": 0},
            {"id": "task2", "route": 1},
            {"id": "task2", "route": 2},
            {"id": "task3", "route": 0},
        ]

        data["sequence"] = copy.deepcopy(task_sequence)
        state = conducting.WorkflowState.deserialize(data)

        expected_task_sequence = [
            (1, {"id": "task2", "route": 0}),
            (2, {"id": "task2", "route": 1}),
            (3, {"id": "task2", "route": 2}),
        ]

        actual_task_sequence = state.get_tasks(task_id="task2", last_occurrence=False)

        self.assertListEqual(actual_task_sequence, expected_task_sequence)

    def test_get_tasks_by_task_id_and_route(self):
        data = copy.deepcopy(MOCK_WORKFLOW_STATE)

        task_sequence = [
            {"id": "task1", "route": 0},
            {"id": "task2", "route": 0},
            {"id": "task2", "route": 0},
            {"id": "task2", "route": 1},
            {"id": "task3", "route": 0},
        ]

        data["sequence"] = copy.deepcopy(task_sequence)
        state = conducting.WorkflowState.deserialize(data)

        expected_task_sequence = [
            (1, {"id": "task2", "route": 0}),
            (2, {"id": "task2", "route": 0}),
        ]

        actual_task_sequence = state.get_tasks(task_id="task2", route=0, last_occurrence=False)

        self.assertListEqual(actual_task_sequence, expected_task_sequence)

        expected_task_sequence = [(3, {"id": "task2", "route": 1})]

        actual_task_sequence = state.get_tasks(task_id="task2", route=1, last_occurrence=False)

        self.assertListEqual(actual_task_sequence, expected_task_sequence)

    def test_get_tasks_by_status(self):
        data = copy.deepcopy(MOCK_WORKFLOW_STATE)

        task_sequence = [
            {"id": "task1", "route": 0, "status": "succeeded"},
            {"id": "task2", "route": 0, "status": "failed"},
            {"id": "task2", "route": 0, "status": "succeeded"},
            {"id": "task3", "route": 0, "status": "running"},
        ]

        data["sequence"] = copy.deepcopy(task_sequence)
        state = conducting.WorkflowState.deserialize(data)

        expected_task_sequence = [
            (0, {"id": "task1", "route": 0, "status": "succeeded"}),
            (2, {"id": "task2", "route": 0, "status": "succeeded"}),
        ]

        actual_task_sequence = state.get_tasks_by_status(statuses.SUCCEEDED, last_occurrence=False)

        self.assertListEqual(actual_task_sequence, expected_task_sequence)

    def test_get_terminal_tasks(self):
        data = copy.deepcopy(MOCK_WORKFLOW_STATE)

        task_sequence = [
            {"id": "task1", "route": 0, "status": "succeeded"},
            {"id": "task2", "route": 0, "status": "succeeded", "term": True},
            {"id": "task3", "route": 0, "status": "succeeded", "term": True},
        ]

        data["sequence"] = copy.deepcopy(task_sequence)
        state = conducting.WorkflowState.deserialize(data)

        expected_task_sequence = [
            (1, {"id": "task2", "route": 0, "status": "succeeded", "term": True}),
            (2, {"id": "task3", "route": 0, "status": "succeeded", "term": True}),
        ]

        actual_task_sequence = state.get_terminal_tasks()

        self.assertListEqual(actual_task_sequence, expected_task_sequence)

    def test_get_tasks_last_occurrence(self):
        data = copy.deepcopy(MOCK_WORKFLOW_STATE)

        task_sequence = [
            {"id": "task1", "route": 0},
            {"id": "task1", "route": 0},
            {"id": "task2", "route": 0},
            {"id": "task3", "route": 0},
            {"id": "task3", "route": 0},
            {"id": "task3", "route": 0},
            {"id": "task4", "route": 0},
            {"id": "task5", "route": 0},
        ]

        data["sequence"] = copy.deepcopy(task_sequence)

        task_map = {"task1__r0": 1, "task2__r0": 2, "task3__r0": 5, "task4__r0": 6, "task5__r0": 7}

        data["tasks"] = copy.deepcopy(task_map)

        state = conducting.WorkflowState.deserialize(data)

        expected_task_sequence = [
            (1, {"id": "task1", "route": 0}),
            (2, {"id": "task2", "route": 0}),
            (5, {"id": "task3", "route": 0}),
            (6, {"id": "task4", "route": 0}),
            (7, {"id": "task5", "route": 0}),
        ]

        actual_task_sequence = state.get_tasks(last_occurrence=True)

        self.assertListEqual(actual_task_sequence, expected_task_sequence)

    def test_get_tasks_last_occurrence_by_task_id(self):
        data = copy.deepcopy(MOCK_WORKFLOW_STATE)

        task_sequence = [
            {"id": "task1", "route": 0},
            {"id": "task2", "route": 0},
            {"id": "task2", "route": 0},
            {"id": "task2", "route": 1},
            {"id": "task2", "route": 1},
            {"id": "task3", "route": 0},
        ]

        data["sequence"] = copy.deepcopy(task_sequence)

        task_map = {"task1__r0": 0, "task2__r0": 2, "task2__r1": 4, "task3__r0": 5}

        data["tasks"] = copy.deepcopy(task_map)

        state = conducting.WorkflowState.deserialize(data)

        expected_task_sequence = [
            (2, {"id": "task2", "route": 0}),
            (4, {"id": "task2", "route": 1}),
        ]

        actual_task_sequence = state.get_tasks(task_id="task2", last_occurrence=True)

        self.assertListEqual(actual_task_sequence, expected_task_sequence)

    def test_get_tasks_last_occurrence_by_task_id_and_route(self):
        data = copy.deepcopy(MOCK_WORKFLOW_STATE)

        task_sequence = [
            {"id": "task1", "route": 0},
            {"id": "task2", "route": 0},
            {"id": "task2", "route": 0},
            {"id": "task2", "route": 1},
            {"id": "task2", "route": 1},
            {"id": "task3", "route": 0},
        ]

        data["sequence"] = copy.deepcopy(task_sequence)

        task_map = {"task1__r0": 0, "task2__r0": 2, "task2__r1": 4, "task3__r0": 5}

        data["tasks"] = copy.deepcopy(task_map)

        state = conducting.WorkflowState.deserialize(data)

        expected_task_sequence = [(2, {"id": "task2", "route": 0})]

        actual_task_sequence = state.get_tasks(task_id="task2", route=0, last_occurrence=True)

        self.assertListEqual(actual_task_sequence, expected_task_sequence)

        expected_task_sequence = [(4, {"id": "task2", "route": 1})]

        actual_task_sequence = state.get_tasks(task_id="task2", route=1, last_occurrence=True)

        self.assertListEqual(actual_task_sequence, expected_task_sequence)

    def test_get_tasks_last_occurrence_by_status(self):
        data = copy.deepcopy(MOCK_WORKFLOW_STATE)

        task_sequence = [
            {"id": "task1", "route": 0, "status": "succeeded"},
            {"id": "task2", "route": 0, "status": "succeeded"},
            {"id": "task2", "route": 0, "status": "succeeded"},
            {"id": "task3", "route": 0, "status": "running"},
        ]

        data["sequence"] = copy.deepcopy(task_sequence)

        task_map = {"task1__r0": 0, "task2__r0": 2, "task3__r0": 3}

        data["tasks"] = copy.deepcopy(task_map)

        state = conducting.WorkflowState.deserialize(data)

        expected_task_sequence = [
            (0, {"id": "task1", "route": 0, "status": "succeeded"}),
            (2, {"id": "task2", "route": 0, "status": "succeeded"}),
        ]

        actual_task_sequence = state.get_tasks_by_status(statuses.SUCCEEDED, last_occurrence=True)

        self.assertListEqual(actual_task_sequence, expected_task_sequence)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six


def merge_dicts(left, right, overwrite=True):
    if left is None:
        return right

    if right is None:
        return left

    for k, v in six.iteritems(right):
        if k not in left:
            left[k] = v
        else:
            left_v = left[k]

            if isinstance(left_v, dict) and isinstance(v, dict):
                merge_dicts(left_v, v, overwrite=overwrite)
            elif overwrite:
                left[k] = v

    return left


def get_dict_value(obj, path, raise_key_error=False):
    item = obj
    traversed = ""

    for key in path.split("."):
        if not isinstance(item, dict) and traversed != path:
            raise TypeError("Value of '%s' is not typeof dict." % traversed)

        traversed += "." + key if len(traversed) > 0 else key

        if key not in item and raise_key_error:
            raise KeyError("Key '%s' does not exist." % traversed)

        item = item.get(key, None)

        if item is None:
            break

    return item


def set_dict_value(obj, path, value, raise_key_error=False, insert_null=True):
    if not insert_null and value is None:
        return

    item = obj
    traversed = ""

    for key in path.split("."):
        if not isinstance(item, dict) and traversed != path:
            raise TypeError("Value of '%s' is not typeof dict." % traversed)

        traversed += "." + key if len(traversed) > 0 else key

        if key not in item and raise_key_error:
            raise KeyError("Key '%s' does not exist." % traversed)

        if traversed == path:
            item[key] = value
        else:
            if key not in item:
                item[key] = {}

            item = item[key]



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six

from orquesta.specs import native as native_specs
from orquesta.tests.unit.specs.native import base as test_base


class TaskSpecTest(test_base.OrchestraWorkflowSpecTest):
    def test_delay_bad_syntax(self):
        wf_def = """
            version: 1.0
            description: A basic workflow with a delay in task.
            tasks:
              task1:
                delay: true
                action: core.noop
        """

        expected_errors = {
            "syntax": [
                {
                    "message": "True is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$." "properties.delay.oneOf"
                    ),
                    "spec_path": "tasks.task1.delay",
                }
            ]
        }

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_delay(self):
        wf_def = """
            version: 1.0
            description: A basic workflow with a delay in task.
            tasks:
              task1:
                delay: 100
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

        task1 = wf_spec.tasks["task1"]

        self.assertEqual(task1.delay, 100)

    def test_delay_with_expression(self):
        wf_def = """
            version: 1.0
            description: A basic workflow with a delay in task.
            vars:
              - delay: 100
            tasks:
              task1:
                delay: <% ctx().delay %>
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

        task1 = wf_spec.tasks["task1"]

        self.assertEqual(task1.delay, "<% ctx().delay %>")

    def test_delay_with_bad_vars(self):
        wf_def = """
            version: 1.0
            description: A basic workflow with a delay in task.
            tasks:
              task1:
                delay: <% ctx().delay %>
                action: core.noop
        """

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx().delay %>",
                    "message": 'Variable "delay" is referenced before assignment.',
                    "schema_path": "properties.tasks.patternProperties.^\\w+$.properties.delay",
                    "spec_path": "tasks.task1.delay",
                }
            ]
        }

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_with_items_bad_syntax(self):
        wf_def = """
            version: 1.0
            description: A basic with items workflow.
            tasks:
              task1:
                with:
                  items: foobar
                action: core.noop
        """

        expected_errors = {
            "syntax": [
                {
                    "message": (
                        "'foobar' does not match '%s'" % native_specs.ItemizedSpec._items_regex
                    ),
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.with.properties.items.pattern"
                    ),
                    "spec_path": "tasks.task1.with.items",
                }
            ]
        }

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_inline_with_items_bad_syntax(self):
        wf_def = """
            version: 1.0
            description: A basic with items workflow.
            tasks:
              task1:
                with: foobar
                action: core.noop
        """

        expected_errors = {
            "syntax": [
                {
                    "message": (
                        "'foobar' does not match '%s'" % native_specs.ItemizedSpec._items_regex
                    ),
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.with.properties.items.pattern"
                    ),
                    "spec_path": "tasks.task1.with.items",
                }
            ]
        }

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_multiple_with_items_bad_syntax(self):
        wf_def = """
            version: 1.0
            description: A basic with items workflow.
            tasks:
              task1:
                with:
                  items: foo; bar in <% zip(list(1, 2), list(a, b)) %>
                action: core.noop
              task2:
                with:
                  items: foo bar in <% zip(list(1, 2), list(a, b)) %>
                action: core.noop
        """

        expected_errors = {
            "syntax": [
                {
                    "message": (
                        "'foo; bar in <%% zip(list(1, 2), list(a, b)) %%>' "
                        "does not match '%s'" % native_specs.ItemizedSpec._items_regex
                    ),
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.with.properties.items.pattern"
                    ),
                    "spec_path": "tasks.task1.with.items",
                },
                {
                    "message": (
                        "'foo bar in <%% zip(list(1, 2), list(a, b)) %%>' "
                        "does not match '%s'" % native_specs.ItemizedSpec._items_regex
                    ),
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.with.properties.items.pattern"
                    ),
                    "spec_path": "tasks.task2.with.items",
                },
            ]
        }

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_multiple_inline_with_items_bad_syntax(self):
        wf_def = """
            version: 1.0
            description: A basic with items workflow.
            tasks:
              task1:
                with: foo; bar in <% zip(list(1, 2), list(a, b)) %>
                action: core.noop
              task2:
                with: foo bar in <% zip(list(1, 2), list(a, b)) %>
                action: core.noop
        """

        expected_errors = {
            "syntax": [
                {
                    "message": (
                        "'foo; bar in <%% zip(list(1, 2), list(a, b)) %%>' "
                        "does not match '%s'" % native_specs.ItemizedSpec._items_regex
                    ),
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.with.properties.items.pattern"
                    ),
                    "spec_path": "tasks.task1.with.items",
                },
                {
                    "message": (
                        "'foo bar in <%% zip(list(1, 2), list(a, b)) %%>' "
                        "does not match '%s'" % native_specs.ItemizedSpec._items_regex
                    ),
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.with.properties.items.pattern"
                    ),
                    "spec_path": "tasks.task2.with.items",
                },
            ]
        }

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_with_one_list(self):
        wf_def = """
            version: 1.0

            description: Send direct y to xs

            input:
              - xs
              - batch_size: 1

            tasks:
              task1:
                with: <% ctx(xs) %>
                action: core.noop
              task2:
                with: "{{ ctx('xs') }}"
                action: core.noop
              task3:
                with:
                  items: <% ctx(xs) %>
                  concurrency: <% ctx(batch_size) %>
                action: core.noop
              task4:
                with:
                  items: "{{ ctx('xs') }}"
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

        expected_items = "<% ctx(xs) %>"
        t1_with_attr = getattr(wf_spec.tasks["task1"], "with")
        self.assertEqual(t1_with_attr.items, expected_items)

        expected_items = "{{ ctx('xs') }}"
        t2_with_attr = getattr(wf_spec.tasks["task2"], "with")
        self.assertEqual(t2_with_attr.items, expected_items)

        expected_items = "<% ctx(xs) %>"
        expected_concurrency = "<% ctx(batch_size) %>"
        t3_with_attr = getattr(wf_spec.tasks["task3"], "with")
        self.assertEqual(t3_with_attr.items, expected_items)
        self.assertEqual(t3_with_attr.concurrency, expected_concurrency)

        expected_items = "{{ ctx('xs') }}"
        t4_with_attr = getattr(wf_spec.tasks["task4"], "with")
        self.assertEqual(t4_with_attr.items, expected_items)

    def test_with_one_list_and_var_expansion(self):
        wf_def = """
            version: 1.0

            description: Send direct y to xs

            input:
              - xs
              - batch_size: 1

            tasks:
              task1:
                with: x in <% ctx(xs) %>
                action: core.noop
              task2:
                with: "x in {{ ctx('xs') }}"
                action: core.noop
              task3:
                with:
                  items: x in <% ctx(xs) %>
                  concurrency: <% ctx(batch_size) %>
                action: core.noop
              task4:
                with:
                  items: "x in {{ ctx('xs') }}"
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

        expected_items = "x in <% ctx(xs) %>"
        t1_with_attr = getattr(wf_spec.tasks["task1"], "with")
        self.assertEqual(t1_with_attr.items, expected_items)

        expected_items = "x in {{ ctx('xs') }}"
        t2_with_attr = getattr(wf_spec.tasks["task2"], "with")
        self.assertEqual(t2_with_attr.items, expected_items)

        expected_items = "x in <% ctx(xs) %>"
        expected_concurrency = "<% ctx(batch_size) %>"
        t3_with_attr = getattr(wf_spec.tasks["task3"], "with")
        self.assertEqual(t3_with_attr.items, expected_items)
        self.assertEqual(t3_with_attr.concurrency, expected_concurrency)

        expected_items = "x in {{ ctx('xs') }}"
        t4_with_attr = getattr(wf_spec.tasks["task4"], "with")
        self.assertEqual(t4_with_attr.items, expected_items)

    def test_with_two_lists(self):
        wf_def = """
            version: 1.0

            description: Send direct customized y to xs

            input:
              - xs
              - ys

            tasks:
              task1:
                with: x, y in <% zip(ctx(xs), ctx(ys)) %>
                action: core.noop
              task2:
                with: "x, y in {{ zip(ctx('xs'), ctx('ys')) }}"
                action: core.noop
              task3:
                with:
                  items: x, y in <% zip(ctx(xs), ctx(ys)) %>
                action: core.noop
              task4:
                with:
                  items: "x, y in {{ zip(ctx('xs'), ctx('ys')) }}"
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

        expected_items = "x, y in <% zip(ctx(xs), ctx(ys)) %>"
        t1_with_attr = getattr(wf_spec.tasks["task1"], "with")
        self.assertEqual(t1_with_attr.items, expected_items)

        expected_items = "x, y in {{ zip(ctx('xs'), ctx('ys')) }}"
        t2_with_attr = getattr(wf_spec.tasks["task2"], "with")
        self.assertEqual(t2_with_attr.items, expected_items)

        expected_items = "x, y in <% zip(ctx(xs), ctx(ys)) %>"
        t3_with_attr = getattr(wf_spec.tasks["task3"], "with")
        self.assertEqual(t3_with_attr.items, expected_items)

        expected_items = "x, y in {{ zip(ctx('xs'), ctx('ys')) }}"
        t4_with_attr = getattr(wf_spec.tasks["task4"], "with")
        self.assertEqual(t4_with_attr.items, expected_items)

    def test_with_many_lists(self):
        wf_def = """
            version: 1.0

            description: Send direct customized y to xs

            input:
              - xs
              - ys
              - zs

            tasks:
              task1:
                with: x, y, z in <% zip(ctx(xs), ctx(ys), ctx(zs)) %>
                action: core.noop
              task2:
                with: "x, y, z in {{ zip(ctx('xs'), ctx('ys'), ctx('zs')) }}"
                action: core.noop
              task3:
                with:
                  items: x, y, z in <% zip(ctx(xs), ctx(ys), ctx(zs)) %>
                action: core.noop
              task4:
                with:
                  items: "x, y, z in {{ zip(ctx('xs'), ctx('ys'), ctx('zs')) }}"
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

        expected_items = "x, y, z in <% zip(ctx(xs), ctx(ys), ctx(zs)) %>"
        t1_with_attr = getattr(wf_spec.tasks["task1"], "with")
        self.assertEqual(t1_with_attr.items, expected_items)

        expected_items = "x, y, z in {{ zip(ctx('xs'), ctx('ys'), ctx('zs')) }}"
        t2_with_attr = getattr(wf_spec.tasks["task2"], "with")
        self.assertEqual(t2_with_attr.items, expected_items)

        expected_items = "x, y, z in <% zip(ctx(xs), ctx(ys), ctx(zs)) %>"
        t3_with_attr = getattr(wf_spec.tasks["task3"], "with")
        self.assertEqual(t3_with_attr.items, expected_items)

        expected_items = "x, y, z in {{ zip(ctx('xs'), ctx('ys'), ctx('zs')) }}"
        t4_with_attr = getattr(wf_spec.tasks["task4"], "with")
        self.assertEqual(t4_with_attr.items, expected_items)

    def test_with_various_spacing(self):
        wf_def = """
            version: 1.0

            description: Send direct y to xs

            input:
              - xs
              - ys

            tasks:
              task1:
                with: " <% ctx(xs) %> "
                action: core.noop
              task2:
                with: "  <% ctx(xs) %>  "
                action: core.noop
              task3:
                with: " x in <% ctx(xs) %> "
                action: core.noop
              task4:
                with: "  x in <% ctx(xs) %>  "
                action: core.noop
              task5:
                with: "x  in  <% ctx(xs) %>"
                action: core.noop
              task6:
                with: "x,y in <% zip(ctx(xs), ctx(ys)) %>"
                action: core.noop
              task7:
                with: "x,  y in <% zip(ctx(xs), ctx(ys)) %>"
                action: core.noop
              task8:
                with:
                  items: " x in <% ctx(xs) %> "
                action: core.noop
              task9:
                with:
                  items: "  x in <% ctx(xs) %>  "
                action: core.noop
              task10:
                with:
                  items: "x  in  <% ctx(xs) %>"
                action: core.noop
              task11:
                with:
                  items: "x,y in <% zip(ctx(xs), ctx(ys)) %>"
                action: core.noop
              task12:
                with:
                  items: "x,  y in <% zip(ctx(xs), ctx(ys)) %>"
                action: core.noop
        """

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), {})

        expected_items = {
            "task1": " <% ctx(xs) %> ",
            "task2": "  <% ctx(xs) %>  ",
            "task3": " x in <% ctx(xs) %> ",
            "task4": "  x in <% ctx(xs) %>  ",
            "task5": "x  in  <% ctx(xs) %>",
            "task6": "x,y in <% zip(ctx(xs), ctx(ys)) %>",
            "task7": "x,  y in <% zip(ctx(xs), ctx(ys)) %>",
            "task8": " x in <% ctx(xs) %> ",
            "task9": "  x in <% ctx(xs) %>  ",
            "task10": "x  in  <% ctx(xs) %>",
            "task11": "x,y in <% zip(ctx(xs), ctx(ys)) %>",
            "task12": "x,  y in <% zip(ctx(xs), ctx(ys)) %>",
        }

        for task_name, items_expr in six.iteritems(expected_items):
            tk_with_attr = getattr(wf_spec.tasks[task_name], "with")
            self.assertEqual(tk_with_attr.items, items_expr)

    def test_with_items_bad_vars(self):
        wf_def = """
            version: 1.0
            description: A basic with items workflow.
            tasks:
              task1:
                with:
                  items: x in <% ctx(xs) %>
                  concurrency: <% ctx(size) %>
                action: core.noop
        """

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "<% ctx(size) %>",
                    "message": 'Variable "size" is referenced before assignment.',
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.with.properties.concurrency"
                    ),
                    "spec_path": "tasks.task1.with.concurrency",
                },
                {
                    "type": "yaql",
                    "expression": "x in <% ctx(xs) %>",
                    "message": 'Variable "xs" is referenced before assignment.',
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.with.properties.items"
                    ),
                    "spec_path": "tasks.task1.with.items",
                },
            ]
        }

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_inline_with_items_bad_vars(self):
        wf_def = """
            version: 1.0
            description: A basic with items workflow.
            tasks:
              task1:
                with: x in <% ctx(xs) %>
                action: core.noop
        """

        expected_errors = {
            "context": [
                {
                    "type": "yaql",
                    "expression": "x in <% ctx(xs) %>",
                    "message": 'Variable "xs" is referenced before assignment.',
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.with.properties.items"
                    ),
                    "spec_path": "tasks.task1.with.items",
                }
            ]
        }

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_retry_basic(self):
        wf_def = """
            version: 1.0
            tasks:
              task1:
                action: core.noop
                retry:
                  when: <% failed() %>
                  count: 3
                  delay: 10
        """

        wf_spec = self.instantiate(wf_def)
        self.assertDictEqual(wf_spec.inspect(), {})

    def test_retry_missing_when_count_delay(self):
        wf_def = """
            version: 1.0
            tasks:
              task1:
                action: core.noop
                retry: {}
        """

        expected_errors = {
            "syntax": [
                {
                    "message": "'count' is a required property",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$." "properties.retry.required"
                    ),
                    "spec_path": "tasks.task1.retry",
                }
            ]
        }

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), expected_errors)

    def test_retry_bad_vars(self):
        wf_def = """
            version: 1.0
            tasks:
              task1:
                action: core.noop
                retry:
                  # should be a string
                  when: 1
                  # should be an int
                  count:
                    - "abc"
                  # should be a number
                  delay: true
        """

        expected_errors = {
            "syntax": [
                {
                    "message": "['abc'] is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.retry.properties.count.oneOf"
                    ),
                    "spec_path": "tasks.task1.retry.count",
                },
                {
                    "message": "True is not valid under any of the given schemas",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.retry.properties.delay.oneOf"
                    ),
                    "spec_path": "tasks.task1.retry.delay",
                },
                {
                    "message": "1 is not of type 'string'",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.retry.properties.when.type"
                    ),
                    "spec_path": "tasks.task1.retry.when",
                },
            ]
        }

        wf_spec = self.instantiate(wf_def)

        self.assertDictEqual(wf_spec.inspect(), expected_errors)



# Copyright 2019 Extreme Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from orquesta import conducting
from orquesta.specs import native as native_specs
from orquesta import statuses
from orquesta.tests.unit import base as test_base


class WorkflowConductorWithItemsTest(test_base.WorkflowConductorWithItemsTest):
    def test_empty_items_list(self):
        wf_def = """
        version: 1.0

        vars:
          - xs: []

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": []}
        task_action_specs = []

        mock_ac_ex_statuses = []
        expected_task_statuses = [statuses.SUCCEEDED]
        expected_workflow_statuses = [statuses.SUCCEEDED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert the workflow output is correct.
        conductor.render_workflow_output()
        expected_output = {"items": []}
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_bad_with_items_syntax(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with:
                items: <% ctx(xs) %>
                action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        expected_errors = {
            "semantics": [
                {
                    "message": "The action property is required for with items task.",
                    "schema_path": "properties.tasks.patternProperties.^\\w+$",
                    "spec_path": "tasks.task1",
                }
            ],
            "syntax": [
                {
                    "message": "Additional properties are not allowed ('action' was unexpected)",
                    "schema_path": (
                        "properties.tasks.patternProperties.^\\w+$."
                        "properties.with.additionalProperties"
                    ),
                    "spec_path": "tasks.task1.with",
                }
            ],
        }

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), expected_errors)

    def test_with_items_that_is_action_less(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with:
                items: <% ctx(xs) %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        expected_errors = {
            "semantics": [
                {
                    "message": "The action property is required for with items task.",
                    "schema_path": "properties.tasks.patternProperties.^\\w+$",
                    "spec_path": "tasks.task1",
                }
            ]
        }

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), expected_errors)

    def test_basic_items_list(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED] * 4
        expected_task_statuses = [statuses.RUNNING] * 3 + [statuses.SUCCEEDED]
        expected_workflow_statuses = [statuses.RUNNING] * 3 + [statuses.SUCCEEDED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert the workflow output is correct.
        conductor.render_workflow_output()
        expected_output = {"items": task_ctx["xs"]}
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_basic_items_list_with_different_types(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - null
              - ""
              - foobar
              - 0
              - 1
              - 123
              - 123.456
              - False
              - True
              - {}
              - {'foobar': 'fubar'}
              - []
              - ['foobar', 'fubar']

        tasks:
          task1:
            with: <% ctx(xs) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"

        task_ctx = {
            "xs": [
                None,
                "",
                "foobar",
                0,
                1,
                123,
                123.456,
                False,
                True,
                {},
                {"foobar": "fubar"},
                [],
                ["foobar", "fubar"],
            ]
        }

        task_action_specs = []
        for idx, item in enumerate(task_ctx["xs"]):
            task_action_specs.append(
                {"action": "core.echo", "input": {"message": item}, "item_id": idx}
            )

        items_count = len(task_ctx["xs"])
        mock_ac_ex_statuses = [statuses.SUCCEEDED] * items_count
        expected_task_statuses = [statuses.RUNNING] * (items_count - 1) + [statuses.SUCCEEDED]
        expected_workflow_statuses = [statuses.RUNNING] * (items_count - 1) + [statuses.SUCCEEDED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert the workflow output is correct.
        conductor.render_workflow_output()
        expected_output = {"items": task_ctx["xs"]}
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_basic_items_list_with_concurrency(self):
        wf_def = """
        version: 1.0

        vars:
          - concurrency: 2
          - xs:
              - fee
              - fi
              - fo
              - fum

        tasks:
          task1:
            with:
              items: <% ctx(xs) %>
              concurrency: <% ctx(concurrency) %>
            action: core.echo message=<% item() %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        concurrency = 2

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["fee", "fi", "fo", "fum"], "concurrency": 2}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "fee"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fi"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "fo"}, "item_id": 2},
            {"action": "core.echo", "input": {"message": "fum"}, "item_id": 3},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED] * 4
        expected_task_statuses = [statuses.RUNNING] * 3 + [statuses.SUCCEEDED]
        expected_workflow_statuses = [statuses.RUNNING] * 3 + [statuses.SUCCEEDED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            task_ctx["xs"],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
            concurrency=concurrency,
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

    def test_multiple_items_list(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - foo
              - fu
              - marco
          - ys:
              - bar
              - bar
              - polo

        tasks:
          task1:
            with: x, y in <% zip(ctx(xs), ctx(ys)) %>
            action: core.echo message=<% item(x) + item(y) %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["foo", "fu", "marco"], "ys": ["bar", "bar", "polo"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "foobar"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fubar"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "marcopolo"}, "item_id": 2},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED] * 3
        expected_task_statuses = [statuses.RUNNING] * 2 + [statuses.SUCCEEDED]
        expected_workflow_statuses = [statuses.RUNNING] * 2 + [statuses.SUCCEEDED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            [i[0] + i[1] for i in zip(task_ctx["xs"], task_ctx["ys"])],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert the workflow output is correct.
        conductor.render_workflow_output()
        expected_output = {"items": ["foobar", "fubar", "marcopolo"]}
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_multiple_items_list_with_different_types(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - null
              - ""
              - foobar
              - 0
              - 1
              - 123
              - 123.456
              - False
              - True
              - {}
              - {'foobar': 'fubar'}
              - []
              - ['foobar', 'fubar']
          - ys:
              - null
              - ""
              - foobar
              - 0
              - 1
              - 123
              - 123.456
              - False
              - True
              - {}
              - {'foobar': 'fubar'}
              - []
              - ['foobar', 'fubar']
        tasks:
          task1:
            with: x, y in <% zip(ctx(xs), ctx(ys)) %>
            action: core.foobar
            input:
              x: <% item(x) %>
              y: <% item(y) %>
            next:
              - publish:
                  - items: <% result() %>

        output:
          - items: <% ctx(items) %>
        """

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"

        task_ctx = {
            "xs": [
                None,
                "",
                "foobar",
                0,
                1,
                123,
                123.456,
                False,
                True,
                {},
                {"foobar": "fubar"},
                [],
                ["foobar", "fubar"],
            ],
            "ys": [
                None,
                "",
                "foobar",
                0,
                1,
                123,
                123.456,
                False,
                True,
                {},
                {"foobar": "fubar"},
                [],
                ["foobar", "fubar"],
            ],
        }

        task_action_specs = []
        for idx, pair in enumerate(zip(task_ctx["xs"], task_ctx["ys"])):
            task_action_specs.append(
                {"action": "core.foobar", "input": {"x": pair[0], "y": pair[1]}, "item_id": idx}
            )

        items_count = len(task_ctx["xs"])
        mock_ac_ex_statuses = [statuses.SUCCEEDED] * items_count
        expected_task_statuses = [statuses.RUNNING] * (items_count - 1) + [statuses.SUCCEEDED]
        expected_workflow_statuses = [statuses.RUNNING] * (items_count - 1) + [statuses.SUCCEEDED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            list(zip(task_ctx["xs"], task_ctx["ys"])),
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)

        # Assert the workflow output is correct.
        conductor.render_workflow_output()
        expected_output = {"items": [[i, j] for i, j in zip(task_ctx["xs"], task_ctx["ys"])]}
        self.assertDictEqual(conductor.get_workflow_output(), expected_output)

    def test_multiple_items_list_with_concurrency(self):
        wf_def = """
        version: 1.0

        vars:
          - xs:
              - foo
              - fu
              - marco
          - ys:
              - bar
              - bar
              - polo

        tasks:
          task1:
            with:
              items: x, y in <% zip(ctx(xs), ctx(ys)) %>
              concurrency: 1
            action: core.echo message=<% item(x) + item(y) %>
        """

        concurrency = 1

        spec = native_specs.WorkflowSpec(wf_def)
        self.assertDictEqual(spec.inspect(), {})

        conductor = conducting.WorkflowConductor(spec)
        conductor.request_workflow_status(statuses.RUNNING)

        # Mock the action execution for each item and assert expected task statuses.
        task_route = 0
        task_name = "task1"
        task_ctx = {"xs": ["foo", "fu", "marco"], "ys": ["bar", "bar", "polo"]}

        task_action_specs = [
            {"action": "core.echo", "input": {"message": "foobar"}, "item_id": 0},
            {"action": "core.echo", "input": {"message": "fubar"}, "item_id": 1},
            {"action": "core.echo", "input": {"message": "marcopolo"}, "item_id": 2},
        ]

        mock_ac_ex_statuses = [statuses.SUCCEEDED] * 3
        expected_task_statuses = [statuses.RUNNING] * 2 + [statuses.SUCCEEDED]
        expected_workflow_statuses = [statuses.RUNNING] * 2 + [statuses.SUCCEEDED]

        self.assert_task_items(
            conductor,
            task_name,
            task_route,
            task_ctx,
            [i[0] + i[1] for i in zip(task_ctx["xs"], task_ctx["ys"])],
            task_action_specs,
            mock_ac_ex_statuses,
            expected_task_statuses,
            expected_workflow_statuses,
            concurrency=concurrency,
        )

        # Assert the task is removed from staging.
        self.assertIsNone(conductor.workflow_state.get_staged_task(task_name, task_route))

        # Assert the workflow succeeded.
        self.assertEqual(conductor.get_workflow_status(), statuses.SUCCEEDED)
